@@ -36,6 +36,8 @@
 #include <X11/keysym.h>
 #include <locale.h>
 
+#include <Imlib2.h>
+
 #ifdef HAVE_LIBCONFIG
 #include <libconfig.h>
 #endif
@@ -50,11 +52,18 @@
 #define BORDER_WIDTH            3
 
 #define GRID_COLOR				"orange"
-//#define DEFAULT_FONT    "-*-courier-bold-*-*-12-*-*-*-*-*-*-*"
-#define DEFAULT_FONT    "fixed"
 
 #define CONFIG_FILE		"/.qworc"
-#define MAX_CONFIG_PATH 30
+#define MAX_CONFIG_PATH 50
+
+#define DATA_PATH		DATADIR "/" PACKAGE_NAME "/"
+
+#define MAX_IMAGE_NAME 10
+#define MAX_IMAGE_PATH 50
+
+#define MAX_IMAGES		3
+
+#define IMAGE_SUFFIXE ".png"
 
 #define MAX_CHAR_PER_REGION		 5
 
@@ -73,10 +82,10 @@
 		(abs(b - a) == 1) ? ( (b - a) == 1) : (b - a) < 0
 
 char charset[][MAX_REGIONS] = {
-		"abc??"">>z",
-		"def?????",
-		"ghijk???",
-		"??l>,???",
+		"a,z?>>>c",
+		"deb?????",
+		"fgihj???",
+		"??l>k???",
 		"??mnopq?",
 		"????tsr?",
 		"????.vuw",
@@ -84,20 +93,22 @@ char charset[][MAX_REGIONS] = {
 };
 
 KeySym char_free[MAX_REGIONS][MAX_REGIONS] = {
-	{XK_KP_5, 0xffff, XK_Up, 0xffff, XK_Right, 0xffff, XK_KP_0, 0xffff, XK_Left},
-	{0xffff, XK_KP_1, XK_Tab, XK_question, 0xffff, 0xffff, 0xffff, 0xffff, XK_slash},
-	{0xffff, 0xffff, XK_KP_2, XK_minus, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff},
-	{0xffff, 0xffff, 0xffff, XK_KP_3, XK_backslash, 0xffff, 0xffff, 0xffff, 0xffff},
-	{0xffff, 0xffff, 0xffff, 0xffff, XK_KP_6, XK_Return, 0xffff, 0xffff, 0xffff},
-	{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, XK_KP_9, 0xffff, 0xffff, 0xffff},
-	{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, XK_underscore, XK_KP_8, 0xffff, 0xffff},
-	{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, XK_KP_7, 0xffff},
-	{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, XK_Control_L, XK_Down, XK_KP_4}};
+	{XK_KP_5, 0xffff, XK_Up, 0xffff, XK_Right, XK_greater, XK_KP_0, XK_less, XK_Left},
+	{0xffff, XK_KP_1, XK_Tab, XK_KP_Multiply, 0xffff, 0xffff, 0xffff, XK_parenleft, XK_slash},
+	{XK_Down, XK_exclam, XK_KP_2, XK_minus, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff},
+	{0xffff, XK_equal, XK_dollar, XK_KP_3, XK_backslash, XK_parenright, 0xffff, 0xffff, 0xffff},
+	{0xffff, 0xffff, 0xffff, XK_plus, XK_KP_6, XK_Return, 0xffff, 0xffff, 0xffff},
+	{0xffff, 0xffff, 0xffff, XK_colon, XK_semicolon, XK_KP_9, XK_bracketright, XK_at, 0xffff},
+	{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, XK_underscore, XK_KP_8, XK_braceright, 0xffff},
+	{0xffff, XK_ampersand, 0xffff, 0xffff, 0xffff, XK_braceleft, XK_bracketleft, XK_KP_7, XK_numbersign},
+	{0xffff, XK_Escape, 0xffff, 0xffff, 0xffff, 0xffff, XK_Control_L, XK_quotedbl, XK_KP_4}};
 
 static KeyCode Shift_code, Control_code;
 
 KeySym custom_charset[MAX_REGIONS - 1][MAX_CHAR_PER_REGION];
 
+char image_names[][MAX_IMAGE_NAME] = { "normal", "caps", "extra" };
+
 const XPoint point1 = { WIDTH / 3, 0 };
 const XPoint point2 = { 2*(WIDTH / 3), 0 };
 const XPoint point3 = { WIDTH, (HEIGHT / 3) };
@@ -124,6 +135,8 @@ const XPoint point_nw = { WIDTH, 0};
 const XPoint point_sw = { WIDTH, HEIGHT};
 const XPoint point_se = { 0, HEIGHT};
 
+static Pixmap char_pixmaps[3];
+
 typedef enum {
 	KeyboardNone = 0,
 	KeyboardShow,
@@ -179,67 +192,7 @@ void init_regions(Display *dpy, Window toplevel)
 	XFree(wm_hints);
 }
 
-int load_font(Display *dpy, XFontStruct **font_info, char *font)
-{
-	if (font == NULL)
-		font = DEFAULT_FONT;
-	if ((*font_info = XLoadQueryFont(dpy, font)) == NULL){
-		fprintf( stderr, "Cannot open font %s\n", font);
-		return 0;
-	}
-	return 1;
-}
-
-void display_charset(Display *dpy, Window win, GC gc, XFontStruct *font_info, int upper_case)
-{
-	int len;
-	int font_height;
-	int i,j, count;
-	int offset;
-	XTextItem item = { NULL, 1, 0, font_info->fid };
-
-	font_height = font_info->ascent + font_info->descent;
-	len = XTextWidth(font_info, &charset[0][0], 1);
-	offset = WIDTH >> 3;
-
-	count = 0;
-	for (i = 0; i < 3; i++) {
-		for (j = 0; j < 3 && count != 8; j++, count++) {
-			char c;
-			if (upper_case) {
-				c = toupper((int) charset[i][j]);
-				item.chars = &c;
-			}
-			else
-				item.chars = &charset[i][j];
-			XDrawText(dpy, win, gc, offset * count, font_height, &item, 1);
-			if (upper_case) {
-				c = toupper((int) charset[i + 4][j + 4]);
-				item.chars = &c;
-			}
-			else
-				item.chars = &charset[i + 4][j + 4];
-			XDrawText(dpy, win, gc, WIDTH - (offset * count) - len, HEIGHT - font_info->descent, &item, 1);
-			if (upper_case) {
-				c = toupper((int) charset[i + 2][j + 2]);
-				item.chars = &c;
-			}
-			else
-				item.chars = &charset[i + 2][j + 2];
-			XDrawText(dpy, win, gc, WIDTH - 2 * len, offset * count + font_height, &item, 1);
-			if (upper_case) {
-				c = toupper((int) charset[(i + 6) & 7][(j + 6) & 7]);
-				item.chars = &c;
-			}
-			else
-				item.chars = &charset[(i + 6) & 7][(j + 6) & 7];
-			XDrawText(dpy, win, gc, len, HEIGHT - (offset * count), &item, 1);
-		}
-	}
-
-}
-
-void draw_grid(Display *dpy, Window toplevel, GC gc)
+void draw_grid(Display *dpy, Pixmap pixmap, GC gc)
 {
 	XColor grid_color, exact;
 	Colormap cmap;
@@ -251,21 +204,67 @@ void draw_grid(Display *dpy, Window toplevel, GC gc)
 
 	XSetForeground(dpy, gc, grid_color.pixel);
 
-	XDrawLine(dpy, toplevel, gc, point1.x, point1.y, point9.x, point9.y);
-	XDrawLine(dpy, toplevel, gc, point2.x, point2.y, point10.x, point10.y);
+	XDrawLine(dpy, pixmap, gc, point1.x, point1.y, point9.x, point9.y);
+	XDrawLine(dpy, pixmap, gc, point2.x, point2.y, point10.x, point10.y);
 
-	XDrawLine(dpy, toplevel, gc, point3.x, point3.y, point11.x, point11.y);
-	XDrawLine(dpy, toplevel, gc, point4.x, point4.y, point12.x, point12.y);
+	XDrawLine(dpy, pixmap, gc, point3.x, point3.y, point11.x, point11.y);
+	XDrawLine(dpy, pixmap, gc, point4.x, point4.y, point12.x, point12.y);
 
-	XDrawLine(dpy, toplevel, gc, point5.x, point5.y, point13.x, point13.y);
-	XDrawLine(dpy, toplevel, gc, point6.x, point6.y, point14.x, point14.y);
+	XDrawLine(dpy, pixmap, gc, point5.x, point5.y, point13.x, point13.y);
+	XDrawLine(dpy, pixmap, gc, point6.x, point6.y, point14.x, point14.y);
+
+	XDrawLine(dpy, pixmap, gc, point7.x, point7.y, point15.x, point15.y);
+	XDrawLine(dpy, pixmap, gc, point8.x, point8.y, point16.x, point16.y);
 
-	XDrawLine(dpy, toplevel, gc, point7.x, point7.y, point15.x, point15.y);
-	XDrawLine(dpy, toplevel, gc, point8.x, point8.y, point16.x, point16.y);
 	XSetForeground(dpy, gc, blackColor);
 
 }
 
+int load_charset(Display *dpy, GC gc, int num){
+	Visual *vis;
+	Colormap cm;
+	int depth;
+	char image_path[MAX_IMAGE_PATH];
+	Imlib_Image image;
+
+	unsigned long blackColor = BlackPixel(dpy, DefaultScreen(dpy));
+	unsigned long whiteColor = WhitePixel(dpy, DefaultScreen(dpy));
+
+	XSetForeground(dpy, gc, whiteColor);
+	XFillRectangle(dpy, char_pixmaps[num], gc, 0, 0, WIDTH, HEIGHT);
+	XSetForeground(dpy, gc, blackColor);
+
+	vis = DefaultVisual(dpy, DefaultScreen(dpy));
+	depth = DefaultDepth(dpy, DefaultScreen(dpy));
+	cm = DefaultColormap(dpy, DefaultScreen(dpy));
+
+	imlib_context_set_display(dpy);
+	imlib_context_set_visual(vis);
+	imlib_context_set_colormap(cm);
+	imlib_context_set_progress_function(NULL);
+
+	imlib_context_set_drawable(char_pixmaps[num]);
+
+	strncpy(image_path, DATA_PATH, MAX_IMAGE_PATH);
+	strncat(image_path + strlen(DATA_PATH), image_names[num],
+		MAX_IMAGE_PATH - strlen(DATA_PATH));
+	strncat(image_path + strlen(DATA_PATH) + strlen(image_names[num]),
+			IMAGE_SUFFIXE, MAX_IMAGE_PATH - strlen(DATA_PATH) - strlen(image_names[num]));
+	image = imlib_load_image(image_path);
+	if (!image) {
+		fprintf(stderr, "%s : Can't open file\n", image_path);
+		return 1;
+	}
+	imlib_context_set_image(image);
+	imlib_render_image_on_drawable(0, 0);
+
+	draw_grid(dpy, char_pixmaps[num], gc);
+
+	imlib_free_image();
+
+	return 0;
+}
+
 #ifdef HAVE_LIBCONFIG
 int read_config(char *config_path)
 {
@@ -363,15 +362,25 @@ int init_keycodes(Display *dpy){
 	return 0;
 }
 
+void update_display(Display *dpy, Window toplevel, GC gc, int shift, int help){
+
+	XClearWindow(dpy, toplevel);
+	if (help) {
+		XCopyArea(dpy, char_pixmaps[MAX_IMAGES - 1], toplevel, gc, 0,0, WIDTH, HEIGHT, 0, 0);
+	} else if (shift) {
+		XCopyArea(dpy, char_pixmaps[1], toplevel, gc, 0,0, WIDTH, HEIGHT, 0, 0);
+	} else {
+		XCopyArea(dpy, char_pixmaps[0], toplevel, gc, 0,0, WIDTH, HEIGHT, 0, 0);
+	}
+	XSync(dpy, False);
+}
+
 int main(int argc, char **argv)
 {
 	Display *dpy;
 	char *display_name;
 	Window toplevel;
-	XSetWindowAttributes attributes;
 
-	char *font_name = NULL;
-	XFontStruct *font_info;
 	unsigned long valuemask;
 	XGCValues xgc;
 	GC gc;
@@ -388,6 +397,8 @@ int main(int argc, char **argv)
 	int buffer_count = 0;
 	int shift_modifier = 0;
 	int ctrl_modifier = 0;
+	int help_screen = 0;
+	int i, status = 0;
 	Time last_cross_timestamp = 0L;
 	Time last_pressed = 0L;
 
@@ -395,9 +406,6 @@ int main(int argc, char **argv)
 	options = getopt(argc, argv, "f:c:");
 
 	switch(options){
-		case 'f':
-			font_name = optarg;
-			break;
 		case 'c':
 			config_path = optarg;
 			break;
@@ -439,22 +447,16 @@ int main(int argc, char **argv)
 
 	XShapeQueryVersion(dpy, &shape_ext_major, &shape_ext_minor);
 
-	unsigned long blackColor = BlackPixel(dpy, DefaultScreen(dpy));
-	unsigned long whiteColor = WhitePixel(dpy, DefaultScreen(dpy));
+	xgc.foreground = BlackPixel(dpy, DefaultScreen(dpy));
+	xgc.background = WhitePixel(dpy, DefaultScreen(dpy));
+	xgc.line_width = 2;
+	valuemask = GCForeground | GCBackground | GCLineWidth;
 
-	attributes.background_pixel = whiteColor;
-	attributes.override_redirect = False;
-	valuemask = CWBackPixel;
+	gc = XCreateGC(dpy, DefaultRootWindow(dpy), valuemask, &xgc);
 
 	toplevel = XCreateWindow(dpy, DefaultRootWindow(dpy), 0, 0,
 			WIDTH, HEIGHT, 0, CopyFromParent, CopyFromParent,
-			CopyFromParent, valuemask, &attributes);
-
-	xgc.foreground = blackColor;
-	xgc.line_width = 2;
-	valuemask = GCForeground | GCLineWidth;
-
-	gc = XCreateGC(dpy, toplevel, valuemask, &xgc);
+			CopyFromParent, 0, NULL);
 
 	init_regions(dpy, toplevel);
 
@@ -466,17 +468,14 @@ int main(int argc, char **argv)
 
 	XMapWindow(dpy, toplevel);
 
-	if(!(load_font(dpy, &font_info, font_name))) {
-		XDestroyWindow(dpy, toplevel);
-		XCloseDisplay(dpy);
-		exit(-1);
+	for( i = 0; i < MAX_IMAGES; i++) {
+		char_pixmaps[i] = XCreatePixmap(dpy, toplevel, WIDTH, HEIGHT, DefaultDepth(dpy, DefaultScreen(dpy)));
+		status |= load_charset(dpy, gc, i);
 	}
+	if (status)
+		run = 0;
 
-	XSetFont(dpy, gc, font_info->fid);
-
-	draw_grid(dpy, toplevel, gc);
-	display_charset(dpy, toplevel, gc, font_info, shift_modifier);
-	XSync(dpy, False);
+	update_display(dpy, toplevel, gc, shift_modifier, help_screen);
 
 	while(run) {
 		XEvent e;
@@ -524,6 +523,7 @@ int main(int argc, char **argv)
 				XFetchName(dpy, e.xcrossing.window, &region_name);
 				region = region_name[0] - 48;
 				XFree(region_name);
+				KeySym index;
 
 				char c = '\0';
 
@@ -531,7 +531,6 @@ int main(int argc, char **argv)
 
 					if ((buffer[1] == buffer[buffer_count - 1]) && loaded_config && (buffer_count > 2)) {
 						buffer_count = (buffer_count - 1) >> 1;
-						KeySym index;
 
 						if (DIRECTION(buffer[1], buffer[2]))
 							index = custom_charset[buffer[1] - 1][(buffer_count << 1) - 1];
@@ -549,10 +548,11 @@ int main(int argc, char **argv)
 						c = charset[buffer[1] - 1][buffer[buffer_count - 1] - 1];
 							// X11 KeySym maps ASCII table
 						code = XKeysymToKeycode(dpy, c);
-						for (state_mod = 0; state_mod < 4; state_mod++) {
-							if (XKeycodeToKeysym(dpy, code, state_mod) == (KeySym) c) {
-								break;
-							}
+						index = (KeySym) c;
+					}
+					for (state_mod = 0; state_mod < 4; state_mod++) {
+						if (XKeycodeToKeysym(dpy, code, state_mod) == index) {
+							break;
 						}
 					}
 
@@ -575,19 +575,19 @@ int main(int argc, char **argv)
 							}
 							XTestFakeKeyEvent(dpy, code, True, 0);
 							XTestFakeKeyEvent(dpy, code, False, 0);
-						} else if (shift_modifier) {
-							shift_modifier = 0;
+						} else if (shift_modifier || help_screen) {
+							shift_modifier = help_screen = 0;
 						} else if (buffer_count == 4) {
 							shift_modifier = 1;
 						} else if (buffer_count == 5) {
 							shift_modifier = 2;
+						} else if (buffer_count == 6) {
+							help_screen = 1;
 						}
 
 						if (buffer_count != 2) {
 							XClearWindow(dpy, toplevel);
-							draw_grid(dpy, toplevel, gc);
-							display_charset(dpy, toplevel, gc, font_info, shift_modifier);
-							XSync(dpy, False);
+							update_display(dpy, toplevel, gc, shift_modifier, help_screen);
 							buffer_count = 1;
 							buffer[0] = 0;
 							break;
@@ -611,9 +611,7 @@ int main(int argc, char **argv)
 						if (shift_modifier == 1) {
 							shift_modifier = 0;
 							XClearWindow(dpy, toplevel);
-							draw_grid(dpy, toplevel, gc);
-							display_charset(dpy, toplevel, gc, font_info, shift_modifier);
-							XSync(dpy, False);
+							update_display(dpy, toplevel, gc, shift_modifier, help_screen);
 						}
 					}
 
@@ -633,17 +631,13 @@ int main(int argc, char **argv)
 			case Expose:
 			case ConfigureNotify:
 				XMapWindow(dpy, toplevel);
-				draw_grid(dpy, toplevel, gc);
-				display_charset(dpy, toplevel, gc, font_info, shift_modifier);
-				XSync(dpy, False);
+				update_display(dpy, toplevel, gc, shift_modifier, help_screen);
 			case ClientMessage:
 				if ((e.xclient.message_type == mb_im_invoker_command) ||
 					(e.xclient.message_type == mtp_im_invoker_command)) {
 					if (e.xclient.data.l[0] == KeyboardShow) {
 						XMapWindow(dpy, toplevel);
-						draw_grid(dpy, toplevel, gc);
-						display_charset(dpy, toplevel, gc, font_info, shift_modifier);
-						XSync(dpy, False);
+						update_display(dpy, toplevel, gc, shift_modifier, help_screen);
 					}
 					if (e.xclient.data.l[0] == KeyboardHide) {
 						XUnmapWindow(dpy, toplevel);
@@ -656,9 +650,7 @@ int main(int argc, char **argv)
 							visible = 0;
 						} else {
 							XMapWindow(dpy, toplevel);
-							draw_grid(dpy, toplevel, gc);
-							display_charset(dpy, toplevel, gc, font_info, shift_modifier);
-							XSync(dpy, False);
+							update_display(dpy, toplevel, gc, shift_modifier, help_screen);
 							visible = 1;
 						}
 					}
@@ -672,8 +664,9 @@ int main(int argc, char **argv)
 			}
 	}
 
-	XFreeFont(dpy, font_info);
 	XFreeGC(dpy, gc);
+	XFreePixmap(dpy, char_pixmaps[0]);
+	XFreePixmap(dpy, char_pixmaps[1]);
 	XDestroyWindow(dpy, toplevel);
 	XCloseDisplay(dpy);
 
